# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2021 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
DataSync for Rattail
"""

from __future__ import unicode_literals, absolute_import

import logging

from sqlalchemy import orm

from rattail.db import Session, model
from rattail.db.util import make_topo_sortkey
from rattail.datasync import DataSyncWatcher, DataSyncConsumer, NewDataSyncImportConsumer
from rattail.config import parse_list


log = logging.getLogger(__name__)


class RattailWatcher(DataSyncWatcher):
    """
    DataSync watcher for Rattail databases.
    """
    prunes_changes = True

    def __init__(self, *args, **kwargs):
        super(RattailWatcher, self).__init__(*args, **kwargs)
        self.engine = self.config.rattail_engines[self.dbkey]
        self.topo_sortkey = make_topo_sortkey(self.config.get_model())

    def get_changes(self, lastrun):
        """
        Checks the :class:`rattail.db.model:Change` table in a Rattail
        database, to see if there are any pending changes for the datasync
        daemon.
        """
        session = Session(bind=self.engine)
        changes = session.query(model.Change).all()
        session.expunge_all()
        session.close()
        if not changes:
            return

        # here we sort our payload types (class names) topologically, to take
        # foreign key dependencies into account etc.  but we start with a
        # simple lexical sort on class name, just for kicks maybe..?
        class_names = sorted(set([c.class_name for c in changes]))
        class_names.sort(key=self.topo_sortkey)

        # now create "final" list of changes, honoring class sort
        # TODO: should we force deletions to be last? or first?
        final = []
        for class_name in class_names:
            for change in [c for c in changes if c.class_name == class_name]:
                final.append((change.uuid,
                              model.DataSyncChange(
                                  payload_type=class_name,
                                  payload_key=change.instance_uuid,
                                  deletion=change.deleted)))

        # when a Product has a price relationship change (e.g. it is
        # given or loses the "current price"), the Product record is
        # the only thing technically updated, but the simple Rattail
        # -> Rattail sync will use an importer which *skips* those
        # price relation fields.  so here we explicitly add changes to
        # the queue, which can force processing of the price relation
        # changes.  and we definitely want these to happen *last*.
        if 'Product' in class_names:
            for change in [c for c in changes
                           if c.class_name == 'Product' and not c.deleted]:
                final.append((None, # no uuid since this need not be pruned
                              model.DataSyncChange(
                                  payload_type='ProductPriceAssociation',
                                  payload_key=change.instance_uuid)))

        return final

    def prune_changes(self, keys):
        deleted = 0
        session = Session(bind=self.engine)
        for key in keys:
            if key: # note that key can sometimes be None
                change = session.query(model.Change).get(key)
                if change:
                    session.delete(change)
                    session.flush()
                    deleted += 1
        session.commit()
        session.close()
        return deleted


class RattailConsumer(DataSyncConsumer):
    """
    DataSync consumer for Rattail databases.
    """
    # Set this to a sequence of model names if you wish to *consume* data
    # *only* for those models (and ignore all others).
    consume_models = None

    # Set this to a sequence of model names if you wish to *ignore* data *only*
    # for those models (and consume all others).
    ignore_models = None

    def __init__(self, *args, **kwargs):
        super(RattailConsumer, self).__init__(*args, **kwargs)
        self.engine = self.config.rattail_engines[self.dbkey]
        self.model = self.get_data_model()
        self.topo_sortkey = make_topo_sortkey(self.model)

    def get_data_model(self):
        """
        Subclasses may override this if they have extended the schema.
        Defaults to ``rattail.db.model``.
        """
        return model

    def process_changes(self, host_session, changes):
        """
        Process changes for a Rattail database.
        """
        # First determine which models are represented in the change set.  Some
        # models may be ignored, depending on the subclass logic etc.
        class_names = set([c.payload_type for c in changes])
        if self.consume_models is not None:
            class_names = [c for c in class_names if c in self.consume_models]
        elif self.ignore_models is not None:
            class_names = [c for c in class_names if c not in self.ignore_models]

        # The order will matter because of table foreign key dependencies.  We
        # start with a lexical sort just for kicks maybe...
        class_names = sorted(class_names)
        class_names.sort(key=self.topo_sortkey)

        session = Session(bind=self.engine)
        if self.runas_username:
            session.set_continuum_user(self.runas_username)

        for class_name in class_names:

            # this consumer does not need to process these changes; see
            # FromRattailToRattailBase for one which can process them
            if class_name == 'ProductPriceAssociation':
                continue

            cls = getattr(self.model, class_name)

            for change in [c for c in changes if c.payload_type == class_name]:
                log.debug("processing {0} for {1} {2}".format(
                    "deletion" if change.deletion else "change",
                    change.payload_type, change.payload_key))

                if change.deletion:
                    instance = session.query(cls).get(change.payload_key)
                    if instance:
                        self.delete_instance(session, instance)
                        session.flush()
                    else:
                        log.warning("could not find instance to delete")

                else: # add/update
                    host_instance = host_session.query(cls).get(change.payload_key)
                    if host_instance:

                        # if processing a pack item, must process its unit item
                        # first.  this is because both may be "new" in which
                        # case the unit item must exist before the pack may
                        # reference it.
                        if class_name == 'Product' and host_instance.is_pack_item():
                            self.merge_instance(session, host_instance.unit)

                        self.merge_instance(session, host_instance)
                        session.flush()
                    else:
                        log.warning("could not find host instance to merge")

        session.commit()
        session.close()

    def merge_instance(self, session, instance):
        """
        Merge the given model instance into the given database session.

        Subclasses may define model-specific methods instead of or in addition
        to overriding this generic one.
        """
        class_name = instance.__class__.__name__.lower()
        merger = getattr(self, 'merge_{0}'.format(class_name), None)
        if merger:
            return merger(session, instance)

        # Nothing special defined, so just do the normal thing.
        return session.merge(instance)

    def delete_instance(self, session, instance):
        """
        Delete the given model instance from the given database session.

        Subclasses may define model-specific methods instead of or in addition
        to overriding this generic one.
        """
        class_name = instance.__class__.__name__.lower()

        deleter = getattr(self, 'delete_{0}'.format(class_name), None)
        if deleter:
            return deleter(session, instance)

        predeleter = getattr(self, 'predelete_{0}'.format(class_name), None)
        if predeleter:
            predeleter(session, instance)

        session.delete(instance)

    def merge_product(self, session, source_product):
        """
        This method is somewhat of a hack, in order to properly handle
        :class:`rattail.db.model.Product` instances and the interdependent
        nature of the related :class:`rattail.db.model.ProductPrice` instances.
        """

        target_product = session.merge(source_product)

        # I'm not 100% sure I understand this correctly, but here's my
        # thinking: First we clear the price relationships in case they've
        # actually gone away; then we re-establish any which are currently
        # valid.

        # Setting the price relationship attributes to ``None`` isn't enough to
        # force the ORM to notice a change, since the UUID field is ultimately
        # what it's watching.  So we must explicitly use that field here.
        target_product.regular_price_uuid = None
        target_product.tpr_price_uuid = None
        target_product.sale_price_uuid = None
        target_product.current_price_uuid = None
        target_product.suggested_price_uuid = None

        # If the source instance has currently valid price relationships, then
        # we re-establish them.  We must merge the source price instance in
        # order to be certain it will exist in the target session, and avoid
        # foreign key errors.  However we *still* must also set the UUID fields
        # because again, the ORM is watching those...  This was noticed to be
        # the source of some bugs where successive database syncs were
        # effectively "toggling" the price relationship.  Setting the UUID
        # field explicitly seems to solve it.
        if source_product.regular_price_uuid:
            target_product.regular_price = session.merge(source_product.regular_price)
            target_product.regular_price_uuid = target_product.regular_price.uuid
        if source_product.tpr_price_uuid:
            target_product.tpr_price = session.merge(source_product.tpr_price)
            target_product.tpr_price_uuid = target_product.tpr_price.uuid
        if source_product.sale_price_uuid:
            target_product.sale_price = session.merge(source_product.sale_price)
            target_product.sale_price_uuid = target_product.sale_price.uuid
        if source_product.current_price_uuid:
            target_product.current_price = session.merge(source_product.current_price)
            target_product.current_price_uuid = target_product.current_price.uuid
        if source_product.suggested_price_uuid:
            target_product.suggested_price = session.merge(source_product.suggested_price)
            target_product.suggested_price_uuid = target_product.suggested_price.uuid

        return target_product

    def predelete_department(self, session, department):

        # Disconnect from all subdepartments.
        q = session.query(model.Subdepartment).filter(
            model.Subdepartment.department == department)
        for subdept in q:
            subdept.department = None

        # Disconnect from all products.
        q = session.query(model.Product).filter(
            model.Product.department == department)
        for product in q:
            product.department = None

    def predelete_subdepartment(self, session, subdepartment):

        # Disconnect from all products.
        q = session.query(model.Product).filter(
            model.Product.subdepartment == subdepartment)
        for product in q:
            product.subdepartment = None

    def predelete_family(self, session, family):

        # Disconnect from all products.
        q = session.query(model.Product).filter(
            model.Product.family == family)
        for product in q:
            product.family = None

    def predelete_vendor(self, session, vendor):

        # Remove all product costs.
        q = session.query(model.ProductCost).filter(
            model.ProductCost.vendor == vendor)
        for cost in q:
            session.delete(cost)

    def predelete_customergroup(self, session, group):

        # Disconnect from all customers.
        q = session.query(model.CustomerGroupAssignment).filter(
            model.CustomerGroupAssignment.group == group)
        for assignment in q:
            session.delete(assignment)


class FromRattailToRattailBase(NewDataSyncImportConsumer):
    """
    Base class for Rattail -> Rattail datasync consumers
    """

    def setup(self):
        super(FromRattailToRattailBase, self).setup()
        self.topo_sortkey = make_topo_sortkey(self.model)
        self.handle_price_xref = self.get_handle_price_xref()

    def get_handle_price_xref(self):
        if hasattr(self, 'handle_price_xref'):
            return self.handle_price_xref

        # TODO: this should default to True, once logic is proven
        return self.config.getbool(
            'rattail.datasync', 'rattail.handle_price_xref',
            default=False)

    def get_importers(self):
        importers = super(FromRattailToRattailBase, self).get_importers()

        # maybe add importer to handle product price reference changes
        if self.get_handle_price_xref() and 'ProductPriceAssociation' not in importers:
            importers['ProductPriceAssociation'] = self.handler.get_importer('ProductPriceAssociation')

        return importers

    def get_host_object(self, session, change):
        if change.payload_type == 'ProductPriceAssociation':
            cls = self.model.Product
        else:
            cls = getattr(self.model, change.payload_type)
        return session.query(cls).get(change.payload_key)


class FromRattailToRattailExportConsumer(FromRattailToRattailBase):
    """
    Export data changes from "local" Rattail to another
    """
    handler_spec = None

    def __init__(self, *args, **kwargs):
        super(FromRattailToRattailExportConsumer, self).__init__(*args, **kwargs)
        self.target_engine = self.config.rattail_engines[self.dbkey]
        self.model = self.config.get_model()

    def make_target_session(self):
        return Session(bind=self.target_engine)

    def process_changes(self, session, changes):
        target_session = self.make_target_session()

        if self.runas_username:
            target_session.set_continuum_user(self.runas_username)

        # update all importers with current sessions
        for importer in self.importers.values():
            importer.host_session = session
            importer.session = target_session

        # note that we no longer sort changes here, since the watcher
        # should have already done that...
        for change in changes:
            self.invoke_importer(session, change)

        target_session.commit()
        target_session.close()


class FromRattailToRattailImportConsumer(FromRattailToRattailBase):
    """
    Import data changes from another Rattail to "local"
    """
    handler_spec = None

    def __init__(self, *args, **kwargs):
        super(FromRattailToRattailImportConsumer, self).__init__(*args, **kwargs)
        self.host_engine = self.config.rattail_engines[self.watcher.dbkey]
        self.local_engine = self.config.rattail_engines[self.dbkey]
        self.model = self.config.get_model()

    def process_changes(self, session, changes):
        self.host_session = Session(bind=self.host_engine)
        local_session = Session(bind=self.local_engine)
        if self.runas_username:
            local_session.set_continuum_user(self.runas_username)

        # update all importers with current sessions
        for importer in self.importers.values():
            importer.host_session = self.host_session
            importer.session = local_session

        # topographically sort changes, i.e. per schema table dependencies.
        # we start with a lexical sort just for kicks maybe...
        class_names = set([c.payload_type for c in changes])
        class_names = sorted(class_names)
        class_names.sort(key=self.topo_sortkey)

        for class_name in class_names:
            cls = getattr(self.model, class_name)
            for change in [c for c in changes if c.payload_type == class_name]:
                self.invoke_importer(session, change)

        local_session.commit()
        local_session.close()
        # TODO: should we ever commit this..?
        # self.host_session.commit()
        self.host_session.close()

    def get_host_object(self, session, change):
        # we use base logic to get the object, but must substitute the
        # session to ensure we read from actual host and not local
        return super(FromRattailToRattailImportConsumer, self).get_host_object(
            self.host_session, change)



# -*- coding: utf-8; -*-
"""fix version discrepancies

Revision ID: a7b286e87f89
Revises: 8f9a960b0c5e
Create Date: 2021-01-23 19:24:03.780728

"""

# revision identifiers, used by Alembic.
revision = 'a7b286e87f89'
down_revision = '8f9a960b0c5e'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types


# NOTE: i apparently just missed these somewhere along the way.  hopefully they
# were the only "gaps" but i don't really know that for sure...  this should
# cover all "grow field X" type migrations though, i believe.

def upgrade():

    # role
    op.alter_column('role_version', 'name', type_=sa.String(length=100))

    # category
    op.alter_column('category_version', 'code', type_=sa.String(length=20))

    # label_profile
    op.alter_column('label_profile_version', 'code', type_=sa.String(length=30))


def downgrade():

    # label_profile
    op.alter_column('label_profile_version', 'code', type_=sa.String(length=3))

    # category
    op.alter_column('category_version', 'code', type_=sa.String(length=8))

    # role
    op.alter_column('role_version', 'name', type_=sa.String(length=25))



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2021 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
App Handler
"""

from __future__ import unicode_literals, absolute_import

import os
# import re
import tempfile

from sqlalchemy import orm

from rattail.util import load_object, pretty_quantity
from rattail.files import temp_path
from rattail.mail import send_email


class AppHandler(object):
    """
    Base class and default implementation for top-level Rattail app handler.

    aka. "the handler to handle all handlers"

    aka. "one handler to bind them all"
    """

    def __init__(self, config):
        self.config = config

    def get_board_handler(self, **kwargs):
        if not hasattr(self, 'board_handler'):
            from rattail.board import get_board_handler
            self.board_handler = get_board_handler(self.config, **kwargs)
        return self.board_handler

    def get_clientele_handler(self, **kwargs):
        if not hasattr(self, 'clientele_handler'):
            from rattail.clientele import get_clientele_handler
            self.clientele_handler = get_clientele_handler(self.config, **kwargs)
        return self.clientele_handler

    def get_employment_handler(self, **kwargs):
        if not hasattr(self, 'employment_handler'):
            from rattail.employment import get_employment_handler
            self.employment_handler = get_employment_handler(self.config, **kwargs)
        return self.employment_handler

    def get_feature_handler(self, **kwargs):
        if not hasattr(self, 'feature_handler'):
            from rattail.features import FeatureHandler
            self.feature_handler = FeatureHandler(self.config, **kwargs)
        return self.feature_handler

    def get_email_handler(self, **kwargs):
        if not hasattr(self, 'email_handler'):
            from rattail.mail import get_email_handler
            self.email_handler = get_email_handler(self.config, **kwargs)
        return self.email_handler

    # TODO: is it helpful to expose this? or more confusing?
    get_mail_handler = get_email_handler

    def get_membership_handler(self, **kwargs):
        """
        Returns a reference to the configured Membership Handler.

        See also :doc:`rattail-manual:base/handlers/other/membership`.
        """
        if not hasattr(self, 'membership_handler'):
            spec = self.config.get('rattail', 'membership.handler',
                                   default='rattail.membership:MembershipHandler')
            factory = load_object(spec)
            self.membership_handler = factory(self.config, **kwargs)
        return self.membership_handler

    def get_people_handler(self, **kwargs):
        """
        Returns a reference to the configured People Handler.

        See also :doc:`rattail-manual:base/handlers/other/people`.
        """
        if not hasattr(self, 'people_handler'):
            spec = self.config.get('rattail', 'people.handler',
                                   default='rattail.people:PeopleHandler')
            factory = load_object(spec)
            self.people_handler = factory(self.config, **kwargs)
        return self.people_handler

    def get_products_handler(self, **kwargs):
        if not hasattr(self, 'products_handler'):
            from rattail.products import get_products_handler
            self.products_handler = get_products_handler(self.config, **kwargs)
        return self.products_handler

    def get_report_handler(self, **kwargs):
        if not hasattr(self, 'report_handler'):
            from rattail.reporting import get_report_handler
            self.report_handler = get_report_handler(self.config, **kwargs)
        return self.report_handler

    def get_session(self, obj):
        """
        Returns the SQLAlchemy session with which the given object is
        associated.  Simple convenience wrapper around
        ``sqlalchemy.orm.object_session()``.
        """
        return orm.object_session(obj)

    def make_session(self, **kwargs):
        """
        Creates and returns a new SQLAlchemy session for the Rattail DB.
        """
        from rattail.db import Session
        return Session(**kwargs)

    def cache_model(self, session, model, **kwargs):
        """
        Convenience method which invokes
        :func:`rattail.db.cache.cache_model()` with the given model
        and keyword arguments.
        """
        from rattail.db import cache
        return cache.cache_model(session, model, **kwargs)

    def make_temp_dir(self, **kwargs):
        """
        Create a temporary directory.  This is mostly a convenience wrapper
        around the built-in ``tempfile.mkdtemp()``.
        """
        if 'dir' not in kwargs:
            workdir = self.config.workdir(require=False)
            if workdir:
                tmpdir = os.path.join(workdir, 'tmp')
                if not os.path.exists(tmpdir):
                    os.makedirs(tmpdir)
                kwargs['dir'] = tmpdir
        return tempfile.mkdtemp(**kwargs)

    def make_temp_file(self, **kwargs):
        """
        Reserve a temporary filename.  This is mostly a convenience wrapper
        around the built-in ``tempfile.mkstemp()``.
        """
        if 'dir' not in kwargs:
            workdir = self.config.workdir(require=False)
            if workdir:
                tmpdir = os.path.join(workdir, 'tmp')
                if not os.path.exists(tmpdir):
                    os.makedirs(tmpdir)
                kwargs['dir'] = tmpdir
        return temp_path(**kwargs)

    def normalize_phone_number(self, number):
        """
        Normalize the given phone number, to a "common" format that
        can be more easily worked with for sync logic etc.
        """
        from rattail.db.util import normalize_phone_number

        return normalize_phone_number(number)

    def phone_number_is_invalid(self, number):
        """
        This method should validate the given phone number string, and if the
        number is *not* considered valid, this method should return the reason
        as string.  If the number is valid, returns ``None``.
        """
        # strip non-numeric chars, and make sure we have 10 left
        normal = self.normalize_phone_number(number)
        if len(normal) != 10:
            return "Phone number must have 10 digits"

    def render_currency(self, value, scale=2, **kwargs):
        """
        Must return a human-friendly display string for the given currency
        value, e.g. ``Decimal('4.20')`` becomes ``"$4.20"``.
        """
        if value is not None:
            if value < 0:
                fmt = "(${{:0,.{}f}})".format(scale)
                return fmt.format(0 - value)
            fmt = "${{:0,.{}f}}".format(scale)
            return fmt.format(value)

    def render_quantity(self, value, **kwargs):
        """
        Return a human-friendly display string for the given quantity
        value, e.g. ``1.000`` becomes ``"1"``.
        """
        return pretty_quantity(value, **kwargs)

    def render_date(self, value, **kwargs):
        """
        Must return a human-friendly display string for the given ``date``
        object.
        """
        if value is not None:
            return value.strftime('%Y-%m-%d')

    def render_datetime(self, value, **kwargs):
        """
        Must return a human-friendly display string for the given ``datetime``
        object.
        """
        if value is not None:
            return value.strftime('%Y-%m-%d %I:%M:%S %p')

    def send_email(self, key, data={}, **kwargs):
        """
        Send an email message of the given type.

        See :func:`rattail.mail.send_email()` for more info.
        """
        send_email(self.config, key, data, **kwargs)


class GenericHandler(object):
    """
    Base class for misc. "generic" feature handlers.

    Most handlers which exist for sake of business logic, should inherit from
    this.
    """

    def __init__(self, config, **kwargs):
        self.config = config
        self.enum = self.config.get_enum()
        self.model = self.config.get_model()
        self.app = self.config.get_app()

    def get_session(self, obj):
        """
        Convenience wrapper around :meth:`AppHandler.get_session()`.
        """
        return self.app.get_session(obj)

    def make_session(self):
        """
        Convenience wrapper around :meth:`AppHandler.make_session()`.
        """
        return self.app.make_session()

    def cache_model(self, session, model, **kwargs):
        """
        Convenience method which invokes :func:`rattail.db.cache.cache_model()`
        with the given model and keyword arguments.
        """
        return self.app.cache_model(session, model, **kwargs)


def make_app(config, **kwargs):
    """
    Create and return the configured :class:`AppHandler` instance.
    """
    spec = config.get('rattail', 'app.handler')
    if spec:
        factory = load_object(spec)
    else:
        factory = AppHandler
    return factory(config, **kwargs)



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2021 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Data Batch Handlers
"""

from __future__ import unicode_literals, absolute_import

import os
import shutil
import datetime
import warnings
import logging

import six
from sqlalchemy import orm

from rattail.db import model, api
from rattail.core import Object
from rattail.gpc import GPC
from rattail.barcodes import upce_to_upca
from rattail.db.cache import cache_model
from rattail.time import localtime, make_utc
from rattail.util import progress_loop, load_object


log = logging.getLogger(__name__)


class BatchHandler(object):
    """
    Base class and partial default implementation for batch handlers.  It is
    expected that all batch handlers will ultimately inherit from this base
    class, therefore it defines the implementation "interface" loosely
    speaking.  Custom batch handlers are welcome to supplement or override this
    as needed, and in fact must do so for certain aspects.
    """
    populate_batches = False
    populate_with_versioning = True

    refresh_with_versioning = True
    repopulate_when_refresh = False

    execute_with_versioning = True

    def __init__(self, config):
        self.config = config
        self.app = self.config.get_app()
        self.enum = config.get_enum()
        self.model = config.get_model()

    @property
    def batch_model_class(self):
        """
        Reference to the data model class of the batch type for which this
        handler is responsible,
        e.g. :class:`~rattail.db.model.batch.labels.LabelBatch`.  Each handler
        must define this (or inherit from one that does).
        """
        raise NotImplementedError("You must set the 'batch_model_class' attribute "
                                  "for class '{}'".format(self.__class__.__name__))

    @property
    def batch_key(self):
        """
        The "batch type key" for the handler, e.g. ``'labels'``.  This is not
        meant to uniquely identify the handler itself, but rather the *type* of
        batch which it handles.  Therefore multiple handlers may be defined
        which share the same ``batch_key`` - although in practice usually each
        app will need only one handler per batch type.

        If the handler doesn't define this, it will be obtained from the
        ``batch_key`` attribute of the :attr:`batch_model_class` attribute.
        """
        return self.batch_model_class.batch_key

    def get_model_title(self):
        return self.batch_model_class.get_model_title()

    def allow_versioning(self, action):
        if action == 'populate':
            return self.populate_with_versioning
        if action == 'refresh':
            return self.refresh_with_versioning
        if action == 'execute':
            return self.execute_with_versioning
        raise NotImplementedError("unknown batch action: {}".format(action))

    def is_mutable(self, batch):
        """
        Returns a boolean indicating whether or not *any* modifications are to
        be allowed for the given batch.  By default this returns ``False`` if
        the batch is already executed, or has been marked complete; otherwise
        returns ``True``.
        """
        if batch.executed:
            return False
        if batch.complete:
            return False
        return True

    def make_basic_batch(self, session, user=None, progress=None, **kwargs):
        """
        Make a new "basic" batch, with no customization beyond what is provided
        by ``kwargs``, which are passed directly to the batch class constructor.

        Note that the new batch instance will be added to the provided session,
        which will also be flushed.

        Callers should use :meth:`make_batch()` instead of this method.
        """
        kwargs.setdefault('rowcount', 0)
        kwargs.setdefault('complete', False)

        # try to provide default creator
        if user and 'created_by' not in kwargs and 'created_by_uuid' not in kwargs:
            kwargs['created_by'] = user

        batch = self.batch_model_class(**kwargs)
        session.add(batch)
        session.flush()
        return batch

    def make_batch(self, session, progress=None, **kwargs):
        """
        Make and return a new batch instance.

        This is the method which callers should use.  It invokes
        :meth:`make_basic_batch()` to actually create the new batch instance,
        and then :meth:`init_batch()` to perform any extra initialization for
        it.  Note that the batch returned will *not* yet be fully populated.
        """
        batch = self.make_basic_batch(session, progress=progress, **kwargs)
        kwargs['session'] = session
        self.init_batch(batch, progress=progress, **kwargs)
        return batch

    def init_batch(self, batch, progress=None, **kwargs):
        """
        Perform extra initialization for the given batch, in whatever way might
        make sense.  Default of course is to do nothing here; handlers are free
        to override as needed.

        Note that initial population of a batch should *not* happen here; see
        :meth:`populate()` for a place to define that logic.
        """

    def make_row(self, **kwargs):
        """
        Make a new row for the batch.  Note however that the row will **not**
        be added to the batch; that should be done with :meth:`add_row()`.

        :returns: A new row object, which does *not* yet belong to any batch.
        """
        return self.batch_model_class.row_class(**kwargs)

    def add_row(self, batch, row):
        """
        Add the given row to the given batch.  This assumes it is a *new* row
        which does not yet belong to any batch.  This logic performs the
        following steps:

        The row is officially added to the batch, and is immediately
        "refreshed" via :meth:`refresh_row()`.

        The row is then examined to see if it has been marked as "removed" by
        the refresh.  If it was *not* removed then the batch's cached
        ``rowcount`` is incremented, and the :meth:`after_add_row()` hook is
        invoked.
        """
        session = orm.object_session(batch)
        with session.no_autoflush:
            batch.data_rows.append(row)
            self.refresh_row(row)
        if not row.removed:
            batch.rowcount = (batch.rowcount or 0) + 1
            self.after_add_row(batch, row)

    def after_add_row(self, batch, row):
        """
        Event hook, called immediately after the given row has been "properly"
        added to the batch.  This is a good place to update totals for the
        batch, to account for the new row, etc.
        """

    def purge_batches(self, session, before=None, before_days=90,
                      dry_run=False, delete_all_data=None,
                      progress=None, **kwargs):
        """
        Purge all batches which were executed prior to a given date.

        :param before: If provided, must be a timezone-aware datetime object.
           If not provided, it will be calculated from the current date, using
           ``before_days``.

        :param before_days: Number of days before the current date, to be used
           as the cutoff date if ``before`` is not specified.

        :param dry_run: Flag indicating that this is a "dry run" and all logic
           involved should be (made) aware of that fact.

        :param delete_all_data: Flag indicating whether *all* data should be
           deleted for each batch being purged.  This flag is passed along to
           :meth:`delete()`; see that for more info.  NOTE: This flag should
           probably be deprecated, but so far has not been...but ``dry_run``
           should be preferred for readability etc.

        :returns: Integer indicating the number of batches purged.
        """
        if delete_all_data and dry_run:
            raise ValueError("You can enable (n)either of `dry_run` or "
                             "`delete_all_data` but both cannot be True")
        delete_all_data = not dry_run

        if not before:
            before = localtime(self.config).date() - datetime.timedelta(days=before_days)
            before = datetime.datetime.combine(before, datetime.time(0))
            before = localtime(self.config, before)

        log.info("will purge '%s' batches, executed before %s",
                 self.batch_key, before.date())

        old_batches = session.query(self.batch_model_class)\
                             .filter(self.batch_model_class.executed < before)\
                             .options(orm.joinedload(self.batch_model_class.data_rows))\
                             .all()
        log.info("found %s batches to purge", len(old_batches))
        result = Object()
        result.purged = 0

        def purge(batch, i):
            self.do_delete(batch, dry_run=dry_run)
            result.purged += 1
            if i % 5 == 0:
                session.flush()

        self.progress_loop(purge, old_batches, progress,
                           message="Purging old batches")

        session.flush()
        if old_batches:
            log.info("%spurged %s '%s' batches",
                     "(would have) " if dry_run else "",
                     result.purged, self.batch_key)
        return result.purged

    @property
    def root_datadir(self):
        """
        The absolute path of the root folder in which data for this particular
        type of batch is stored.  The structure of this path is as follows:

        .. code-block:: none

           /{root_batch_data_dir}/{batch_type_key}

        * ``{root_batch_data_dir}`` - Value of the 'batch.files' option in the
          [rattail] section of config file.
        * ``{batch_type_key}`` - Unique key for the type of batch it is.

        .. note::
           While it is likely that the data folder returned by this method
           already exists, this method does not guarantee it.
        """
        return self.config.batch_filedir(self.batch_key)

    def datadir(self, batch):
        """
        Returns the absolute path of the folder in which the batch's source
        data file(s) resides.  Note that the batch must already have been
        persisted to the database.  The structure of the path returned is as
        follows:

        .. code-block:: none

           /{root_datadir}/{uuid[:2]}/{uuid[2:]}

        * ``{root_datadir}`` - Value returned by :meth:`root_datadir()`.
        * ``{uuid[:2]}`` - First two characters of batch UUID.
        * ``{uuid[2:]}`` - All batch UUID characters *after* the first two.

        .. note::
           While it is likely that the data folder returned by this method
           already exists, this method does not guarantee any such thing.  It
           is typically assumed that the path will have been created by a
           previous call to :meth:`make_batch()` however.
        """
        return os.path.join(self.root_datadir, batch.uuid[:2], batch.uuid[2:])

    def make_datadir(self, batch):
        """
        Returns the data folder specific to the given batch, creating if necessary.
        """
        datadir = self.datadir(batch)
        if not os.path.exists(datadir):
            os.makedirs(datadir)
        return datadir

    # TODO: remove default attr?
    def set_input_file(self, batch, path, attr='filename'):
        """
        Assign the data file found at ``path`` to the batch.  This overwrites
        the given attribute (``attr``) of the batch and places a copy of the
        data file in the batch's data folder.
        """
        datadir = self.make_datadir(batch)
        filename = os.path.basename(path)
        shutil.copyfile(path, os.path.join(datadir, filename))
        setattr(batch, attr, filename)

    def should_populate(self, batch):
        """
        Must return a boolean indicating whether the given batch should be
        populated from an initial data source, i.e. at time of batch creation.
        Override this method if you need to inspect the batch in order to
        determine whether the populate step is needed.  Default behavior is to
        simply return the value of :attr:`populate_batches`.
        """
        return self.populate_batches

    def setup_populate(self, batch, progress=None):
        """
        Perform any setup (caching etc.) necessary for populating a batch.
        """

    def teardown_populate(self, batch, progress=None):
        """
        Perform any teardown (cleanup etc.) necessary after populating a batch.
        """

    def do_populate(self, batch, user, progress=None):
        """
        Perform initial population for the batch, i.e. fill it with data rows.
        Where the handler obtains the data to do this, will vary greatly.

        Note that callers *should* use this method, but custom batch handlers
        should *not* override this method.  Conversely, custom handlers
        *should* override the :meth:`~populate()` method, but callers should
        *not* use that one directly.
        """
        self.setup_populate(batch, progress=progress)
        self.populate(batch, progress=progress)
        self.teardown_populate(batch, progress=progress)
        self.refresh_batch_status(batch)
        return True

    def populate(self, batch, progress=None):
        """
        Populate the batch with initial data rows.  It is assumed that the data
        source to be used will be known by inspecting various properties of the
        batch itself.

        Note that callers should *not* use this method, but custom batch
        handlers *should* override this method.  Conversely, custom handlers
        should *not* override the :meth:`~do_populate()` method, but callers
        *should* use that one directly.
        """
        raise NotImplementedError("Please implement `{}.populate()` method".format(batch.__class__.__name__))

    def refreshable(self, batch):
        """
        This method should return a boolean indicating whether or not the
        handler supports a "refresh" operation for the batch, given its current
        condition.  The default assumes a refresh is allowed unless the batch
        has already been executed.
        """
        if batch.executed:
            return False
        return True

    def progress_loop(self, *args, **kwargs):
        return progress_loop(*args, **kwargs)

    def setup_refresh(self, batch, progress=None):
        """
        Perform any setup (caching etc.) necessary for refreshing a batch.
        """

    def teardown_refresh(self, batch, progress=None):
        """
        Perform any teardown (cleanup etc.) necessary after refreshing a batch.
        """

    def do_refresh(self, batch, user, progress=None):
        """
        Perform a full data refresh for the batch, i.e. update any data which
        may have become stale, etc.

        Note that callers *should* use this method, but custom batch handlers
        should *not* override this method.  Conversely, custom handlers
        *should* override the :meth:`~refresh()` method, but callers should
        *not* use that one directly.
        """
        self.refresh(batch, progress=progress)
        return True

    def refresh(self, batch, progress=None):
        """
        Perform a full data refresh for the batch.  What exactly this means will
        depend on the type of batch, and specific handler logic.

        Generally speaking this refresh is meant to use queries etc. to obtain
        "fresh" data for the batch (header) and all its rows.  In most cases
        certain data is expected to be "core" to the batch and/or rows, and
        such data will be left intact, with all *other* data values being
        re-calculated and/or reset etc.

        Note that callers should *not* use this method, but custom batch
        handlers *should* override this method.  Conversely, custom handlers
        should *not* override the :meth:`~do_refresh()` method, but callers
        *should* use that one directly.
        """
        session = orm.object_session(batch)
        self.setup_refresh(batch, progress=progress)
        if self.repopulate_when_refresh:
            del batch.data_rows[:]
            batch.rowcount = 0
            session.flush()
            self.populate(batch, progress=progress)
        else:
            batch.rowcount = 0

            def refresh(row, i):
                with session.no_autoflush:
                    self.refresh_row(row)
                if not row.removed:
                    batch.rowcount += 1

            self.progress_loop(refresh, batch.active_rows(), progress,
                               message="Refreshing batch data rows")
        self.refresh_batch_status(batch)
        self.teardown_refresh(batch, progress=progress)
        return True

    def refresh_many(self, batches, user=None, progress=None):
        """
        Refresh a set of batches, with given progress.  Default behavior is to
        simply refresh each batch in succession.  Any batches which are already
        executed are skipped.

        Handlers may have to override this method if "grouping" or other
        special behavior is needed.
        """
        needs_refresh = [batch for batch in batches
                         if not batch.executed]
        if not needs_refresh:
            return

        # TODO: should perhaps try to make the progress indicator reflect the
        # "total" number of rows across all batches being refreshed?  seems
        # like that might be more accurate, for the user.  but also harder.

        for batch in needs_refresh:
            self.do_refresh(batch, user, progress=progress)

    def refresh_row(self, row):
        """
        This method will be passed a row object which has already been properly
        added to a batch, and which has basic required fields already
        populated.  This method is then responsible for further populating all
        applicable fields for the row, based on current data within the
        appropriate system(s).

        Note that in some cases this method may be called multiple times for
        the same row, e.g. once when first creating the batch and then later
        when a user explicitly refreshes the batch.  The method logic must
        account for this possibility.
        """

    def refresh_product_basics(self, row):
        """
        This method will refresh the "basic" product info for a row.  It
        assumes that the row is derived from
        :class:`~rattail.db.model.batch.core.ProductBatchRowMixin` and that
        ``row.product`` is already set to a valid product.
        """
        product = getattr(row, 'product', None)
        if not product:
            return

        row.item_id = product.item_id
        row.upc = product.upc
        row.brand_name = six.text_type(product.brand or "")
        row.description = product.description
        row.size = product.size

        department = product.department
        row.department_number = department.number if department else None
        row.department_name = department.name if department else None

        subdepartment = product.subdepartment
        row.subdepartment_number = subdepartment.number if subdepartment else None
        row.subdepartment_name = subdepartment.name if subdepartment else None

    def quick_entry(self, session, batch, entry):
        """
        Handle a "quick entry" value, e.g. from user input.  Most frequently this
        value would represent a UPC or similar "ID" value for e.g. a product record,
        and the handler's duty would be to either locate a corresponding row within
        the batch (if one exists), or else add a new row to the batch.

        In any event this method can be customized and in fact has no default
        behavior, so must be defined by a handler.

        :param session: Database sesssion.

        :param batch: Batch for which the quick entry is to be handled.  Note that
           this batch is assumed to belong to the given ``session``.

        :param entry: String value to be handled.  This is generally assumed to
           be from user input (e.g. UPC scan field) but may not always be.

        :returns: New or existing "row" object, for the batch.
        """
        raise NotImplementedError

    def locate_product_for_entry(self, session, entry, **kwargs):
        """
        This method aims to provide sane default logic for locating a
        :class:`~rattail.db.model.products.Product` record for the given
        "entry" - which is assumed to be a string, i.e. "as-is" from whatever
        data source.

        Note that any batch handler is free to override this, and certainly
        some batches will never need it anyway.  The goal is more to establish
        a good API pattern, but the handler logic itself can vary.

        Also, this default logic will try to honor the "configured" product key
        field, and prefer that when attempting the lookup.
        """
        # don't bother if we're given empty "entry" value
        if not entry:
            return

        # try to locate product by uuid before other, more specific key
        # if kwargs.get('try_uuid', True):
        product = session.query(model.Product).get(entry)
        if product:
            return product

        # prefer caller-provided key over configured key
        product_key = kwargs.get('product_key')
        if not product_key:
            product_key = self.config.product_key()

        if product_key == 'upc':

            if entry.isdigit():

                # we first assume the user entry *does* include check digit
                provided = GPC(entry, calc_check_digit=False)
                product = api.get_product_by_upc(session, provided)
                if product:
                    return product

                # but we can also calculate a check digit and try that
                checked = GPC(entry, calc_check_digit='upc')
                product = api.get_product_by_upc(session, checked)
                if product:
                    return product

                # one last trick is to expand UPC-E to UPC-A and then reattempt
                # the lookup, *with* check digit (since it would be known)
                if len(entry) in (6, 8):
                    checked = GPC(upce_to_upca(entry), calc_check_digit='upc')
                    product = api.get_product_by_upc(session, checked)
                    if product:
                        return product

        elif product_key == 'item_id':

            # try to locate product by item_id
            product = api.get_product_by_item_id(session, entry)
            if product:
                return product

        elif product_key == 'scancode':

            # try to locate product by scancode
            product = api.get_product_by_scancode(session, entry)
            if product:
                return product

        # if we made it this far, lookup by product key failed.

        # okay then, let's maybe attempt lookup by "alternate" code
        if kwargs.get('lookup_by_code'):
            product = api.get_product_by_code(session, entry)
            if product:
                return product

    def remove_row(self, row):
        """
        Remove the given row from its batch, and update the batch accordingly.
        How exactly the row is "removed" is up to this method.  Default is to
        set the row's ``removed`` flag, then invoke the
        :meth:`refresh_batch_status()` method.

        Note that callers should *not* use this method, but custom batch
        handlers *should* override this method.  Conversely, custom handlers
        should *not* override the :meth:`do_remove_row()` method, but callers
        *should* use that one directly.
        """
        batch = row.batch
        row.removed = True
        self.refresh_batch_status(batch)

    def do_remove_row(self, row):
        """
        Remove the given row from its batch, and update the batch accordingly.
        Uses the following logic:

        If the row's ``removed`` flag is already set, does nothing and returns
        immediately.

        Otherwise, it invokes :meth:`remove_row()` and then decrements the
        batch ``rowcount`` attribute.

        Note that callers *should* use this method, but custom batch handlers
        should *not* override this method.  Conversely, custom handlers
        *should* override the :meth:`remove_row()` method, but callers should
        *not* use that one directly.
        """
        if row.removed:
            return
        self.remove_row(row)
        batch = row.batch
        if batch.rowcount is not None:
            batch.rowcount -= 1

    def refresh_batch_status(self, batch):
        """
        Update the batch status, as needed.  This method does nothing by
        default, but may be overridden if the overall batch status needs to be
        updated according to the status of its rows.  This method may be
        invoked whenever rows are added, removed, updated etc.
        """

    def write_worksheet(self, batch, progress=None):
        """
        Write a worksheet file, to be downloaded by the user.  Must return the
        file path.
        """
        raise NotImplementedError("Please define logic for `{}.write_worksheet()`".format(
            self.__class__.__name__))

    def update_from_worksheet(self, batch, path, progress=None):
        """
        Save the given file to a batch-specific location, then update the
        batch data from the file contents.
        """
        raise NotImplementedError("Please define logic for `{}.update_from_worksheet()`".format(
            self.__class__.__name__))

    def mark_complete(self, batch, progress=None):
        """
        Mark the given batch as "complete".  This usually is just a matter of
        setting the :attr:`~rattail.db.model.batch.BatchMixin.complete` flag
        for the batch, with the idea that this should "freeze" the batch so
        that another user can verify its state before finally executing it.

        Each handler is of course free to expound on this idea, or to add extra
        logic to this "event" of marking a batch complete.
        """
        batch.complete = True

    def mark_incomplete(self, batch, progress=None):
        """
        Mark the given batch as "incomplete" (aka. pending).  This usually is
        just a matter of clearing the
        :attr:`~rattail.db.model.batch.BatchMixin.complete` flag for the batch,
        with the idea that this should "thaw" the batch so that it may be
        further updated, i.e. it's not yet ready to execute.

        Each handler is of course free to expound on this idea, or to add extra
        logic to this "event" of marking a batch incomplete.
        """
        batch.complete = False

    def why_not_execute(self, batch):
        """
        This method should inspect the given batch and, if there is a reason
        that execution should *not* be allowed for it, the method should return
        a text string indicating that reason.  It should return ``None`` if no
        such reason could be identified, and execution should be allowed.

        Note that it is assumed the batch has not already been executed, since
        execution is globally prevented for such batches.  In other words you
        needn't check for that as a possible reason not to execute.
        """

    def executable(self, batch):
        """
        This method should return a boolean indicating whether or not execution
        should be allowed for the batch, given its current condition.

        While you may override this method, you are encouraged to override
        :meth:`why_not_execute()` instead.  Default logic for this method is as
        follows:

        If the batch is ``None`` then the caller simply wants to know if "any"
        batch may be executed, so we return ``True``.

        If the batch has already been executed then we return ``False``.

        If the :meth:`why_not_execute()` method returns a value, then we assume
        execution is not allowed and return ``False``.

        Finally we will return ``True`` if none of the above rules matched.
        """
        if batch is None:
            return True
        if batch.executed:
            return False
        if self.why_not_execute(batch):
            return False
        return True

    def auto_executable(self, batch):
        """
        Must return a boolean indicating whether the given bath is eligible for
        "automatic" execution, i.e. immediately after batch is created.
        """
        return False

    def describe_execution(self, batch, **kwargs):
        """
        This method should essentially return some text describing briefly what
        will happen when the given batch is executed.

        :param batch: The batch in question, which is a candidate for
           execution.

        :returns: String value describing the batch execution.
        """

    def do_execute(self, batch, user, progress=None, **kwargs):
        """
        Perform final execution for the batch.  What that means for any given
        batch, will vary greatly.

        Note that callers *should* use this method, but custom batch handlers
        should *not* override this method.  Conversely, custom handlers
        *should* override the :meth:`~execute()` method, but callers should
        *not* use that one directly.
        """
        # make sure we declare who's responsible, if we can
        session = orm.object_session(batch)
        session.set_continuum_user(user)

        result = self.execute(batch, user=user, progress=progress, **kwargs)
        if not result:
            return False
        batch.executed = make_utc()
        batch.executed_by = user
        return result

    def execute(self, batch, progress=None, **kwargs):
        """
        Execute the given batch, according to the given kwargs.  This is really
        where the magic happens, although each handler must define that magic,
        since the default logic does nothing at all.

        Note that callers should *not* use this method, but custom batch
        handlers *should* override this method.  Conversely, custom handlers
        should *not* override the :meth:`~do_execute()` method, but callers
        *should* use that one directly.
        """

    def execute_many(self, batches, progress=None, **kwargs):
        """
        Execute a set of batches, with given progress and kwargs.  Default
        behavior is to simply execute each batch in succession.  Any batches
        which are already executed are skipped.

        Handlers may have to override this method if "grouping" or other
        special behavior is needed.
        """
        now = make_utc()
        for batch in batches:
            if not batch.executed:
                self.execute(batch, progress=progress, **kwargs)
                batch.executed = now
                batch.executed_by = kwargs['user']
        return True

    def do_delete(self, batch, dry_run=False, progress=None, **kwargs):
        """
        Totally delete the given batch.  This includes deleting the batch
        itself, any rows and "extra" data such as files.

        Note that callers *should* use this method, but custom batch handlers
        should *not* override this method.  Conversely, custom handlers
        *should* override the :meth:`~delete()` method, but callers should
        *not* use that one directly.
        """
        session = orm.object_session(batch)

        if 'delete_all_data' in kwargs:
            warnings.warn("The 'delete_all_data' kwarg is not supported for "
                          "this method; please use 'dry_run' instead",
                          DeprecationWarning)
        kwargs['delete_all_data'] = not dry_run

        self.delete(batch, progress=progress, **kwargs)
        session.delete(batch)

    def delete(self, batch, delete_all_data=True, progress=None, **kwargs):
        """
        Delete all data for the batch, including any related (e.g. row)
        records, as well as files on disk etc.  This method should *not* delete
        the batch itself however.

        Note that callers should *not* use this method, but custom batch
        handlers *should* override this method.  Conversely, custom handlers
        should *not* override the :meth:`~do_delete()` method, but callers
        *should* use that one directly.

        :param delete_all_data: Flag indicating whether *all* data should be
           deleted.  You should probably set this to ``False`` if in dry-run
           mode, since deleting *all* data often implies deleting files from
           disk, which is not transactional and therefore can't be rolled back.
        """
        if delete_all_data:
            self.delete_extra_data(batch, progress=progress)

        # delete all rows from batch, one by one.  maybe would be nicer if we
        # could delete all in one fell swoop, but sometimes "extension" row
        # records might exist, and can get FK constraint errors
        # TODO: in other words i don't even know why this is necessary.  seems
        # to me that one fell swoop should not incur FK errors
        if hasattr(batch, 'data_rows'):
            session = orm.object_session(batch)

            def delete(row, i):
                session.delete(row)
                if i % 200 == 0:
                    session.flush()

            self.progress_loop(delete, batch.data_rows, progress,
                               message="Deleting rows from batch")
            session.flush()

            # even though we just deleted all rows, we must also "remove" all
            # rows explicitly from the batch; otherwise when the batch itself
            # is deleted, SQLAlchemy may complain about an unexpected number of
            # rows being deleted
            del batch.data_rows[:]

    def delete_extra_data(self, batch, progress=None, **kwargs):
        """
        Delete all "extra" data for the batch.  This method should *not* bother
        trying to delete the batch itself, or rows thereof.  It typically is
        only concerned with deleting extra files on disk, related to the batch.
        """
        path = self.config.batch_filepath(self.batch_key, batch.uuid)
        if os.path.exists(path):
            shutil.rmtree(path)

    def setup_clone(self, oldbatch, progress=None):
        """
        Perform any setup (caching etc.) necessary for cloning batch.  Note
        that the ``oldbatch`` arg is the "old" batch, i.e. the one from which a
        clone is to be created.
        """

    def teardown_clone(self, newbatch, progress=None):
        """
        Perform any teardown (cleanup etc.) necessary after cloning a batch.
        Note that the ``newbatch`` arg is the "new" batch, i.e. the one which
        was just created by cloning the old batch.
        """

    def clone(self, oldbatch, created_by, progress=None):
        """
        Clone the given batch as a new batch, and return the new batch.
        """
        self.setup_clone(oldbatch, progress=progress)
        batch_class = self.batch_model_class
        batch_mapper = orm.class_mapper(batch_class)

        newbatch = batch_class()
        newbatch.created_by = created_by
        newbatch.rowcount = 0
        for name in batch_mapper.columns.keys():
            if name not in ('uuid', 'id', 'created', 'created_by_uuid', 'rowcount', 'executed', 'executed_by_uuid'):
                setattr(newbatch, name, getattr(oldbatch, name))

        session = orm.object_session(oldbatch)
        session.add(newbatch)
        session.flush()

        row_class = newbatch.row_class
        row_mapper = orm.class_mapper(row_class)

        def clone_row(oldrow, i):
            newrow = self.clone_row(oldrow)
            self.add_row(newbatch, newrow)

        self.progress_loop(clone_row, oldbatch.data_rows, progress,
                           message="Cloning data rows for new batch")

        self.refresh_batch_status(newbatch)
        self.teardown_clone(newbatch, progress=progress)
        return newbatch

    def clone_row(self, oldrow):
        row_class = self.batch_model_class.row_class
        row_mapper = orm.class_mapper(row_class)
        newrow = row_class()
        for name in row_mapper.columns.keys():
            if name not in ('uuid', 'batch_uuid', 'sequence'):
                setattr(newrow, name, getattr(oldrow, name))
        return newrow

    def cache_model(self, session, model, **kwargs):
        return cache_model(session, model, **kwargs)


def get_batch_types(config):
    """
    Returns the list of available batch type keys.
    """
    model = config.get_model()

    keys = []
    for name in dir(model):
        if name == 'BatchMixin':
            continue
        obj = getattr(model, name)
        if isinstance(obj, type):
            if issubclass(obj, model.Base):
                if issubclass(obj, model.BatchMixin):
                    keys.append(obj.batch_key)

    keys.sort()
    return keys


def get_batch_handler(config, batch_key, default=None, error=True):
    """
    Returns a batch handler object corresponding to the given batch key.

    Note that this returns the "default" handler for the given batch type, of
    which there can only be one, determined by config.  If you need a specific
    handler then you should create it yourself; this function is useful when
    you just want the default handler.
    """
    # spec is assumed to come from config/settings if present, otherwise
    # caller-supplied default is assumed
    spec = config.get('rattail.batch', '{}.handler'.format(batch_key),
                      default=default)

    # if none of the above gave us a spec, check for common 'importer' type
    if not spec and batch_key == 'importer':
        spec = 'rattail.batch.importer:ImporterBatchHandler'

    if error and not spec:
        raise ValueError("handler spec not found for batch type: {}".format(
            batch_key))

    handler = load_object(spec)(config)
    return handler



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2021 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Models for customer order batches
"""

from __future__ import unicode_literals, absolute_import

import sqlalchemy as sa
from sqlalchemy import orm
from sqlalchemy.ext.declarative import declared_attr

from rattail.db.model import (Base, BatchMixin, BatchRowMixin,
                              CustomerOrderBase, CustomerOrderItemBase,
                              CustomerOrder, CustomerOrderItem)


class CustomerOrderBatch(BatchMixin, CustomerOrderBase, Base):
    """
    Hopefully generic batch used for entering new customer orders into the
    system, as well as fulfilling them along the way, etc.
    """
    batch_key = 'custorder'
    __tablename__ = 'batch_custorder'
    __batchrow_class__ = 'CustomerOrderBatchRow'
    model_title_plural = "Customer Order Batches"

    @declared_attr
    def __table_args__(cls):
        return cls.__batch_table_args__() + cls.__customer_order_table_args__() + (
            sa.ForeignKeyConstraint(['order_uuid'], ['custorder.uuid'],
                                    name='batch_custorder_fk_order'),
        )

    STATUS_OK                           = 1

    STATUS = {
        STATUS_OK                       : "ok",
    }

    order_uuid = sa.Column(sa.String(length=32), nullable=True)
    order = orm.relationship(
        CustomerOrder,
        doc="""
        Reference to the customer order with which the batch is associated.
        May be null, e.g. in the case of a "new order" batch.
        """,
        backref=orm.backref(
            'batches',
            order_by='CustomerOrderBatch.id',
            doc="""
            List of batches associated with the customer order.
            """))

    mode = sa.Column(sa.Integer(), nullable=False, doc="""
    Numeric "mode" for the customer order batch, to indicate new/fulfilling etc.
    """)


class CustomerOrderBatchRow(BatchRowMixin, CustomerOrderItemBase, Base):
    """
    Row of data within a customer order batch.
    """
    __tablename__ = 'batch_custorder_row'
    __batch_class__ = CustomerOrderBatch

    @declared_attr
    def __table_args__(cls):
        return cls.__batchrow_table_args__() + cls.__customer_order_item_table_args__() + (
            sa.ForeignKeyConstraint(['item_uuid'], ['custorder_item.uuid'],
                                    name='batch_custorder_row_fk_item'),
        )

    STATUS_OK                           = 1
    STATUS_PRODUCT_NOT_FOUND            = 2

    STATUS = {
        STATUS_OK                       : "ok",
        STATUS_PRODUCT_NOT_FOUND        : "product not found",
    }

    item_entry = sa.Column(sa.String(length=32), nullable=True, doc="""
    Raw entry value, as obtained from the initial data source, and which is
    used to locate the product within the system.  This raw value is preserved
    in case the initial lookup fails and a refresh must attempt further
    lookup(s) later.  Only used by certain batch handlers in practice.
    """)

    item_uuid = sa.Column(sa.String(length=32), nullable=True)
    item = orm.relationship(
        CustomerOrderItem,
        doc="""
        Reference to the customer order line item with which the batch row is
        associated.  May be null, e.g. in the case of a "new order" batch.
        """)



# -*- coding: utf-8 -*-
"""add purchase_batch.complete

Revision ID: 8c17754a6bbf
Revises: 09976b73e96a
Create Date: 2016-11-08 11:48:52.222366

"""

from __future__ import unicode_literals

# revision identifiers, used by Alembic.
revision = '8c17754a6bbf'
down_revision = u'09976b73e96a'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types
from sqlalchemy.dialects import postgresql


batch_id_seq = sa.Sequence('batch_id_seq')


def upgrade():

    # purchase_batch
    op.add_column('purchase_batch', sa.Column('complete', sa.Boolean(), nullable=True))


def downgrade():

    # purchase_batch
    op.drop_column('purchase_batch', 'complete')



# -*- coding: utf-8 -*-
"""add receiving.truck_dump

Revision ID: c37154504ae1
Revises: d87aeb7da072
Create Date: 2018-05-14 12:17:21.347122

"""

from __future__ import unicode_literals, absolute_import

# revision identifiers, used by Alembic.
revision = 'c37154504ae1'
down_revision = u'd87aeb7da072'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # purchase_batch
    op.add_column('purchase_batch', sa.Column('truck_dump', sa.Boolean(), nullable=True))


def downgrade():

    # purchase_batch
    op.drop_column('purchase_batch', 'truck_dump')



# -*- coding: utf-8; -*-
"""initial Trainwreck tables

Revision ID: 4638ceca90d6
Revises: 
Create Date: 2021-02-15 15:02:18.976439

"""

from __future__ import unicode_literals, absolute_import

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '4638ceca90d6'
down_revision = None
branch_labels = ('trainwreck',)
depends_on = None


def upgrade():

    # transaction
    op.create_table('transaction',
                    sa.Column('uuid', sa.String(length=32), nullable=False),
                    sa.Column('system', sa.String(length=50), nullable=True),
                    sa.Column('system_id', sa.String(length=50), nullable=True),
                    sa.Column('terminal_id', sa.String(length=20), nullable=True),
                    sa.Column('receipt_number', sa.String(length=20), nullable=True),
                    sa.Column('start_time', sa.DateTime(), nullable=True),
                    sa.Column('end_time', sa.DateTime(), nullable=True),
                    sa.Column('upload_time', sa.DateTime(), nullable=True),
                    sa.Column('cashier_id', sa.String(length=20), nullable=True),
                    sa.Column('cashier_name', sa.String(length=255), nullable=True),
                    sa.Column('customer_id', sa.String(length=20), nullable=True),
                    sa.Column('customer_name', sa.String(length=255), nullable=True),
                    sa.Column('shopper_id', sa.String(length=20), nullable=True),
                    sa.Column('shopper_name', sa.String(length=255), nullable=True),
                    sa.Column('subtotal', sa.Numeric(precision=9, scale=2), nullable=True),
                    sa.Column('discounted_subtotal', sa.Numeric(precision=9, scale=2), nullable=True),
                    sa.Column('tax', sa.Numeric(precision=8, scale=2), nullable=True),
                    sa.Column('tax1', sa.Numeric(precision=8, scale=2), nullable=True),
                    sa.Column('tax2', sa.Numeric(precision=8, scale=2), nullable=True),
                    sa.Column('tax3', sa.Numeric(precision=8, scale=2), nullable=True),
                    sa.Column('tax4', sa.Numeric(precision=8, scale=2), nullable=True),
                    sa.Column('tax5', sa.Numeric(precision=8, scale=2), nullable=True),
                    sa.Column('tax6', sa.Numeric(precision=8, scale=2), nullable=True),
                    sa.Column('tax7', sa.Numeric(precision=8, scale=2), nullable=True),
                    sa.Column('cashback', sa.Numeric(precision=9, scale=2), nullable=True),
                    sa.Column('total', sa.Numeric(precision=9, scale=2), nullable=True),
                    sa.Column('void', sa.Boolean(), nullable=False),
                    sa.PrimaryKeyConstraint('uuid')
    )
    op.create_index('transaction_ix_end_time', 'transaction', ['end_time'], unique=False)
    op.create_index('transaction_ix_receipt_number', 'transaction', ['receipt_number'], unique=False)
    op.create_index('transaction_ix_start_time', 'transaction', ['start_time'], unique=False)
    op.create_index('transaction_ix_upload_time', 'transaction', ['upload_time'], unique=False)

    # transaction_item
    op.create_table('transaction_item',
                    sa.Column('uuid', sa.String(length=32), nullable=False),
                    sa.Column('transaction_uuid', sa.String(length=32), nullable=False),
                    sa.Column('sequence', sa.Integer(), nullable=False),
                    sa.Column('item_type', sa.String(length=50), nullable=True),
                    sa.Column('item_scancode', sa.String(length=20), nullable=False),
                    sa.Column('item_id', sa.String(length=50), nullable=True),
                    sa.Column('department_number', sa.Integer(), nullable=True),
                    sa.Column('department_name', sa.String(length=30), nullable=True),
                    sa.Column('exempt_from_gross_sales', sa.Boolean(), nullable=True),
                    sa.Column('brand_name', sa.String(length=100), nullable=True),
                    sa.Column('description', sa.String(length=255), nullable=False),
                    sa.Column('unit_price', sa.Numeric(precision=8, scale=2), nullable=True),
                    sa.Column('unit_regular_price', sa.Numeric(precision=8, scale=2), nullable=True),
                    sa.Column('unit_discounted_price', sa.Numeric(precision=8, scale=2), nullable=True),
                    sa.Column('unit_quantity', sa.Numeric(precision=10, scale=4), nullable=True),
                    sa.Column('subtotal', sa.Numeric(precision=9, scale=2), nullable=True),
                    sa.Column('discounted_subtotal', sa.Numeric(precision=9, scale=2), nullable=True),
                    sa.Column('gross_sales', sa.Numeric(precision=9, scale=2), nullable=True),
                    sa.Column('net_sales', sa.Numeric(precision=9, scale=2), nullable=True),
                    sa.Column('tax', sa.Numeric(precision=8, scale=2), nullable=True),
                    sa.Column('tax1', sa.Numeric(precision=8, scale=2), nullable=True),
                    sa.Column('tax2', sa.Numeric(precision=8, scale=2), nullable=True),
                    sa.Column('tax3', sa.Numeric(precision=8, scale=2), nullable=True),
                    sa.Column('tax4', sa.Numeric(precision=8, scale=2), nullable=True),
                    sa.Column('tax5', sa.Numeric(precision=8, scale=2), nullable=True),
                    sa.Column('tax6', sa.Numeric(precision=8, scale=2), nullable=True),
                    sa.Column('tax7', sa.Numeric(precision=8, scale=2), nullable=True),
                    sa.Column('total', sa.Numeric(precision=9, scale=2), nullable=True),
                    sa.Column('void', sa.Boolean(), nullable=False),
                    sa.Column('error_correct', sa.Boolean(), nullable=False),
                    sa.ForeignKeyConstraint(['transaction_uuid'], ['transaction.uuid'], name='transaction_item_fk_transaction'),
                    sa.PrimaryKeyConstraint('uuid')
    )
    op.create_index('transaction_item_ix_item_id', 'transaction_item', ['item_id'], unique=False)
    op.create_index('transaction_item_ix_transaction_uuid', 'transaction_item', ['transaction_uuid'], unique=False)


def downgrade():

    # transaction_item
    op.drop_index('transaction_item_ix_transaction_uuid', table_name='transaction_item')
    op.drop_index('transaction_item_ix_item_id', table_name='transaction_item')
    op.drop_table('transaction_item')

    # transaction
    op.drop_index('transaction_ix_upload_time', table_name='transaction')
    op.drop_index('transaction_ix_start_time', table_name='transaction')
    op.drop_index('transaction_ix_receipt_number', table_name='transaction')
    op.drop_index('transaction_ix_end_time', table_name='transaction')
    op.drop_table('transaction')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2020 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Base data model for "shopfoo" pattern
"""

from __future__ import unicode_literals, absolute_import

import six
import sqlalchemy as sa
from sqlalchemy import orm
from sqlalchemy.ext.declarative import declared_attr

from rattail.db import model
from rattail.db.core import filename_column


class ShopfooProductBase(object):
    """
    Product cache table, specific to the target system.  Each record in this
    table *should* match exactly, what is in the actual target system.
    """
    @declared_attr
    def __table_args__(cls):
        return cls.__product_table_args__()

    @classmethod
    def __product_table_args__(cls):
        table_name = cls.__tablename__
        return (
            sa.ForeignKeyConstraint(['product_uuid'], ['product.uuid'],
                                    name='{}_fk_product'.format(table_name)),
        )

    uuid = model.uuid_column()

    product_uuid = sa.Column(sa.String(length=32), nullable=True)

    @declared_attr
    def product(cls):
        backref_name = cls.__tablename__
        return orm.relationship(
            model.Product,
            doc="""
            Reference to the actual product record, with which this cache
            record is associated.
            """,
            backref=orm.backref(
                backref_name,
                uselist=False,
                doc="""
                Reference to the local/cached {} record for this product.
                """.format(backref_name)))


@six.python_2_unicode_compatible
class ShopfooProductExportBase(model.ExportMixin):
    """
    Product export table, specific to the target system.  Each record in this
    table will correspond to a single export of product data to the target
    system.
    """
    filename = filename_column(nullable=True, doc="""
    Base filename for the export.  This name should be the same as is used for
    the final upload, if applicable.
    """)

    uploaded = sa.Column(sa.Boolean(), nullable=False, default=False, doc="""
    Flag to indicate whether the export has been successfully uploaded to the
    target system's server.
    """)

    def set_filename(self, config):
        """
        Set the export's official filename, per config.
        """
        table_name = self.__tablename__
        self.filename = "{}.{}.csv".format(table_name, self.id_str)

    def __str__(self):
        return self.id_str



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2021 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Common email config objects
"""

from __future__ import unicode_literals, absolute_import

import sys
import socket
from traceback import format_exception

import six

from rattail.mail import Email
from rattail.util import load_object
from rattail.core import Object
from rattail.time import make_utc, localtime
from rattail.problems import ProblemReport, get_problem_report_handler


class ProblemReportEmail(Email):
    """
    Base class for all "problem report" emails
    """
    abstract = True

    def obtain_sample_data(self, request):
        data = self.sample_data(request)
        handler = get_problem_report_handler(self.config)

        if 'report' not in data:
            reports = handler.get_all_problem_reports()
            email_key = self.__class__.__name__
            for report in reports:
                if report.email_key == email_key:
                    data['report'] = report(self.config)
                    break

            if 'report' not in data:
                report = ProblemReport(self.config)
                report.problem_title = "Generic Title (problem report not found)"
                data['report'] = report

        if 'system_title' not in data:
            system_key = data['report'].system_key or 'rattail'
            data['system_title'] = handler.get_system_title(system_key)

        return data


class datasync_error_watcher_get_changes(Email):
    """
    When any datasync watcher thread encounters an error trying to get changes,
    this email is sent out.
    """
    default_subject = "Watcher failed to get changes"

    def sample_data(self, request):
        from rattail.datasync import DataSyncWatcher
        try:
            raise RuntimeError("Fake error for preview")
        except:
            exc_type, exc, traceback = sys.exc_info()
        watcher = DataSyncWatcher(self.config, 'test')
        watcher.consumes_self = True
        return {
            'watcher': watcher,
            'error': exc,
            'traceback': ''.join(format_exception(exc_type, exc, traceback)).strip(),
            'datasync_url': '/datasyncchanges',
            'attempts': 2,
        }


class datasync_error_consumer_process_changes(Email):
    """
    When any datasync consumer thread encounters an error trying to process
    changes, this email is sent out.
    """
    default_subject = "Consumer failed to process changes"

    def sample_data(self, request):
        from rattail.datasync import DataSyncWatcher, DataSyncConsumer

        try:
            raise RuntimeError("Fake error for preview")
        except:
            exc_type, exc, traceback = sys.exc_info()

        watcher = DataSyncWatcher(self.config, 'testwatcher')
        consumer = DataSyncConsumer(self.config, 'testconsumer')
        return {
            'watcher': watcher,
            'consumer': consumer,
            'error': exc,
            'attempts': 2,
            'traceback': ''.join(format_exception(exc_type, exc, traceback)).strip(),
            'datasync_url': '/datasync/changes',
        }


class filemon_action_error(Email):
    """
    When any filemon thread encounters an error (and the retry attempts have
    been exhausted) then it will send out this email.
    """
    default_subject = "Error invoking action(s)"

    def sample_data(self, request):
        from rattail.filemon import Action
        action = Action(self.config)
        action.spec = 'rattail.filemon.actions:Action'
        action.retry_delay = 10
        try:
            raise RuntimeError("Fake error for preview")
        except:
            exc_type, exc, traceback = sys.exc_info()
        return {
            'hostname': socket.gethostname(),
            'path': '/tmp/foo.csv',
            'action': action,
            'attempts': 3,
            'error': exc,
            'traceback': ''.join(format_exception(exc_type, exc, traceback)).strip(),
        }


class ImporterEmail(Email):
    """
    Sent when a "version catch-up" import is performed, which involves changes.
    """
    abstract = True
    fallback_key = 'rattail_import_updates'
    handler_spec = None

    def get_handler(self, config):
        return load_object(self.handler_spec)(config)

    def get_default_subject(self, **data):
        host_title = data.get('host_title')
        local_title = data.get('local_title')
        if not (host_title and local_title):
            handler = self.get_handler(self.config)
            if not host_title:
                host_title = handler.host_title
            if not local_title:
                local_title = handler.local_title
        return "Changes for {} -> {}".format(host_title, local_title)

    def sample_data(self, request):
        handler = self.get_handler(request.rattail_config)
        obj = Object()
        local_data = {
            'foo': 42,
            'bar': True,
            'baz': 'something',
        }
        host_data = {
            'foo': 42,
            'bar': False,
            'baz': 'something else',
        }
        return {
            'host_title': handler.host_title,
            'local_title': handler.local_title,
            'direction': 'import',
            'runtime': "1 second",
            'argv': ['bin/rattail', 'import-something'],
            'changes': {
                'Widget': (
                    [],                             # created
                    [(obj, local_data, host_data)], # updated
                    [],                             # deleted
                ),
            },
            'render_record': lambda x: six.text_type(x),
            'max_display': 15,
        }


class rattail_export_rattail_updates(ImporterEmail):
    """
    Sent when a full Rattail -> Rattail data export involves changes.
    """
    handler_spec = 'rattail.importing.rattail:FromRattailToRattailExport'
    abstract = False


class rattail_import_rattail_updates(ImporterEmail):
    """
    Sent when a full Rattail -> Rattail data import involves changes.
    """
    handler_spec = 'rattail.importing.rattail:FromRattailToRattailImport'
    abstract = False


class rattail_import_versions_updates(ImporterEmail):
    """
    Sent when a "version catch-up" import is performed, which involves changes.
    """
    handler_spec = 'rattail.importing.versions:FromRattailToRattailVersions'
    abstract = False


class person_merge_request(Email):
    """
    Sent when a request is submitted to merge two Person records.
    """
    default_subject = "Person Merge Request"

    def sample_data(self, request):
        return {
            'user_display': request.user.display_name,
            'removing_display': "Fred Flintstone",
            'removing_url': '#',
            'keeping_display': "Fred Flintstone",
            'keeping_url': '#',
            'merge_request_url': '#',
        }


class upgrade_failure(Email):
    """
    Sent when an app upgrade is attempted, but ultimately failed.
    """
    default_subject = "Upgrade Failure"

    def sample_data(self, request):
        upgrade = Object(
            description="upgrade to the latest!",
            notes="nothing special",
            executed=make_utc(),
            executed_by="Fred Flintstone",
            exit_code=42,
        )
        return {
            # TODO: first 3 args should be provided by some common logic instead...
            'rattail_config': self.config,
            'localtime': localtime,
            'app_title': self.config.app_title(),
            'upgrade': upgrade,
            'upgrade_url': '#',
        }


class upgrade_success(Email):
    """
    Sent when an app upgrade is performed successfully.
    """
    default_subject = "Upgrade Success"

    def sample_data(self, request):
        upgrade = Object(
            description="upgrade to the latest!",
            notes="nothing special",
            executed=make_utc(),
            executed_by="Fred Flintstone",
            exit_code=0,
        )
        return {
            # TODO: first 3 args should be provided by some common logic instead...
            'rattail_config': self.config,
            'localtime': localtime,
            'app_title': self.config.app_title(),
            'upgrade': upgrade,
            'upgrade_url': '#',
        }


class user_feedback(Email):
    """
    Sent when a user submits a Feedback form from the web UI.
    """
    default_subject = "User Feedback"

    def sample_data(self, request):
        return {
            'app_title': request.rattail_config.app_title(default="Rattail"),
            'user': None,
            'user_name': "Fred Flintstone",
            'referrer': request.route_url('home'),
            'client_ip': '127.0.0.1',
            'message': "Hey there,\n\njust wondered what the heck was going on with this site?  It's crap.\n\nFred",
        }



# -*- coding: utf-8 -*-
"""rename tables

Revision ID: 33b76c3cb892
Revises: 2e3f33b8ea46
Create Date: 2015-02-07 12:52:16.186363

"""

from __future__ import unicode_literals

# revision identifiers, used by Alembic.
revision = '33b76c3cb892'
down_revision = u'2e3f33b8ea46'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # Drop all constraints which will be changing.
    op.drop_constraint('roles_uq_name', 'roles', type_='unique')
    op.drop_constraint('permissions_fk_role', 'permissions', type_='foreignkey')
    op.drop_constraint('users_fk_person', 'users', type_='foreignkey')
    op.drop_constraint('users_uq_username', 'users', type_='unique')
    op.drop_constraint('users_roles_fk_role', 'users_roles', type_='foreignkey')
    op.drop_constraint('users_roles_fk_user', 'users_roles', type_='foreignkey')
    op.drop_constraint('employees_fk_person', 'employees', type_='foreignkey')
    op.drop_constraint('customers_groups_fk_group', 'customers_groups', type_='foreignkey')
    op.drop_constraint('customers_groups_fk_customer', 'customers_groups', type_='foreignkey')
    op.drop_constraint('customers_people_fk_customer', 'customers_people', type_='foreignkey')
    op.drop_constraint('customers_people_fk_person', 'customers_people', type_='foreignkey')
    op.drop_constraint('vendors_uq_id', 'vendors', type_='unique')
    op.drop_constraint('vendor_contacts_fk_vendor', 'vendor_contacts', type_='foreignkey')
    op.drop_constraint('vendor_contacts_fk_person', 'vendor_contacts', type_='foreignkey')
    op.drop_constraint('categories_fk_department', 'categories', type_='foreignkey')
    op.drop_constraint('report_codes_uq_code', 'report_codes', type_='unique')
    op.drop_constraint('products_fk_department', 'products', type_='foreignkey')
    op.drop_constraint('products_fk_subdepartment', 'products', type_='foreignkey')
    op.drop_constraint('products_fk_category', 'products', type_='foreignkey')
    op.drop_constraint('products_fk_family', 'products', type_='foreignkey')
    op.drop_constraint('products_fk_report_code', 'products', type_='foreignkey')
    op.drop_constraint('products_fk_brand', 'products', type_='foreignkey')
    op.drop_constraint('products_regular_price_uuid_fkey', 'products', type_='foreignkey')
    op.drop_constraint('products_current_price_uuid_fkey', 'products', type_='foreignkey')
    op.drop_index('products_ix_upc', 'products')
    op.drop_constraint('product_codes_fk_product', 'product_codes', type_='foreignkey')
    op.drop_constraint('product_costs_fk_vendor', 'product_costs', type_='foreignkey')
    op.drop_constraint('product_costs_fk_product', 'product_costs', type_='foreignkey')
    op.drop_constraint('product_prices_fk_product', 'product_prices', type_='foreignkey')
    op.drop_constraint('batch_columns_fk_batch', 'batch_columns', type_='foreignkey')

    # Rename all tables.
    op.rename_table('settings', 'setting')
    op.rename_table('changes', 'change')
    op.rename_table('phone_numbers', 'phone')
    op.rename_table('email_addresses', 'email')
    op.rename_table('people', 'person')
    op.rename_table('roles', 'role')
    op.rename_table('permissions', 'permission')
    op.rename_table('users', 'user')
    op.rename_table('users_roles', 'user_x_role')
    op.rename_table('stores', 'store')
    op.rename_table('employees', 'employee')
    op.rename_table('customer_groups', 'customer_group')
    op.rename_table('customers', 'customer')
    op.rename_table('customers_groups', 'customer_x_group')
    op.rename_table('customers_people', 'customer_x_person')
    op.rename_table('vendors', 'vendor')
    op.rename_table('vendor_contacts', 'vendor_contact')
    op.rename_table('departments', 'department')
    op.rename_table('subdepartments', 'subdepartment')
    op.rename_table('categories', 'category')
    op.rename_table('families', 'family')
    op.rename_table('report_codes', 'report_code')
    op.rename_table('brands', 'brand')
    op.rename_table('products', 'product')
    op.rename_table('product_codes', 'product_code')
    op.rename_table('product_costs', 'product_cost')
    op.rename_table('product_prices', 'product_price')
    op.rename_table('batches', 'batch')
    op.rename_table('batch_columns', 'batch_column')
    op.rename_table('label_profiles', 'label_profile')

    # Recreate all constraints.
    op.create_unique_constraint('role_uq_name', 'role', ['name'])
    op.create_foreign_key('permission_fk_role', 'permission', 'role', ['role_uuid'], ['uuid'])
    op.create_foreign_key('user_fk_person', 'user', 'person', ['person_uuid'], ['uuid'])
    op.create_unique_constraint('user_uq_username', 'user', ['username'])
    op.create_foreign_key('user_x_role_fk_role', 'user_x_role', 'role', ['role_uuid'], ['uuid'])
    op.create_foreign_key('user_x_role_fk_user', 'user_x_role', 'user', ['user_uuid'], ['uuid'])
    op.create_foreign_key('employee_fk_person', 'employee', 'person', ['person_uuid'], ['uuid'])
    op.create_foreign_key('customer_x_group_fk_group', 'customer_x_group', 'customer_group', ['group_uuid'], ['uuid'])
    op.create_foreign_key('customer_x_group_fk_customer', 'customer_x_group', 'customer', ['customer_uuid'], ['uuid'])
    op.create_foreign_key('customer_x_person_fk_customer', 'customer_x_person', 'customer', ['customer_uuid'], ['uuid'])
    op.create_foreign_key('customer_x_person_fk_person', 'customer_x_person', 'person', ['person_uuid'], ['uuid'])
    op.create_unique_constraint('vendor_uq_id', 'vendor', ['id'])
    op.create_foreign_key('vendor_contact_fk_vendor', 'vendor_contact', 'vendor', ['vendor_uuid'], ['uuid'])
    op.create_foreign_key('vendor_contact_fk_person', 'vendor_contact', 'person', ['person_uuid'], ['uuid'])
    op.create_foreign_key('category_fk_department', 'category', 'department', ['department_uuid'], ['uuid'])
    op.create_unique_constraint('report_code_uq_code', 'report_code', ['code'])
    op.create_foreign_key('product_fk_department', 'product', 'department', ['department_uuid'], ['uuid'])
    op.create_foreign_key('product_fk_subdepartment', 'product', 'subdepartment', ['subdepartment_uuid'], ['uuid'])
    op.create_foreign_key('product_fk_category', 'product', 'category', ['category_uuid'], ['uuid'])
    op.create_foreign_key('product_fk_family', 'product', 'family', ['family_uuid'], ['uuid'])
    op.create_foreign_key('product_fk_report_code', 'product', 'report_code', ['report_code_uuid'], ['uuid'])
    op.create_foreign_key('product_fk_brand', 'product', 'brand', ['brand_uuid'], ['uuid'])
    op.create_foreign_key('product_fk_regular_price', 'product', 'product_price', ['regular_price_uuid'], ['uuid'])
    op.create_foreign_key('product_fk_current_price', 'product', 'product_price', ['current_price_uuid'], ['uuid'])
    op.create_index('product_ix_upc', 'product', ['upc'])
    op.create_foreign_key('product_code_fk_product', 'product_code', 'product', ['product_uuid'], ['uuid'])
    op.create_foreign_key('product_cost_fk_vendor', 'product_cost', 'vendor', ['vendor_uuid'], ['uuid'])
    op.create_foreign_key('product_cost_fk_product', 'product_cost', 'product', ['product_uuid'], ['uuid'])
    op.create_foreign_key('product_price_fk_product', 'product_price', 'product', ['product_uuid'], ['uuid'])
    op.create_foreign_key('batch_column_fk_batch', 'batch_column', 'batch', ['batch_uuid'], ['uuid'])


def downgrade():

    # drop all constraints which will be changing.
    op.drop_constraint('role_uq_name', 'role', type_='unique')
    op.drop_constraint('permission_fk_role', 'permission', type_='foreignkey')
    op.drop_constraint('user_fk_person', 'user', type_='foreignkey')
    op.drop_constraint('user_uq_username', 'user', type_='unique')
    op.drop_constraint('user_x_role_fk_role', 'user_x_role', type_='foreignkey')
    op.drop_constraint('user_x_role_fk_user', 'user_x_role', type_='foreignkey')
    op.drop_constraint('employee_fk_person', 'employee', type_='foreignkey')
    op.drop_constraint('customer_x_group_fk_group', 'customer_x_group', type_='foreignkey')
    op.drop_constraint('customer_x_group_fk_customer', 'customer_x_group', type_='foreignkey')
    op.drop_constraint('customer_x_person_fk_customer', 'customer_x_person', type_='foreignkey')
    op.drop_constraint('customer_x_person_fk_person', 'customer_x_person', type_='foreignkey')
    op.drop_constraint('vendor_uq_id', 'vendor', type_='unique')
    op.drop_constraint('vendor_contact_fk_vendor', 'vendor_contact', type_='foreignkey')
    op.drop_constraint('vendor_contact_fk_person', 'vendor_contact', type_='foreignkey')
    op.drop_constraint('category_fk_department', 'category', type_='foreignkey')
    op.drop_constraint('report_code_uq_code', 'report_code', type_='unique')
    op.drop_constraint('product_fk_department', 'product', type_='foreignkey')
    op.drop_constraint('product_fk_subdepartment', 'product', type_='foreignkey')
    op.drop_constraint('product_fk_category', 'product', type_='foreignkey')
    op.drop_constraint('product_fk_family', 'product', type_='foreignkey')
    op.drop_constraint('product_fk_report_code', 'product', type_='foreignkey')
    op.drop_constraint('product_fk_brand', 'product', type_='foreignkey')
    op.drop_constraint('product_fk_regular_price', 'product', type_='foreignkey')
    op.drop_constraint('product_fk_current_price', 'product', type_='foreignkey')
    op.drop_index('product_ix_upc', 'product')
    op.drop_constraint('product_code_fk_product', 'product_code', type_='foreignkey')
    op.drop_constraint('product_cost_fk_vendor', 'product_cost', type_='foreignkey')
    op.drop_constraint('product_cost_fk_product', 'product_cost', type_='foreignkey')
    op.drop_constraint('product_price_fk_product', 'product_price', type_='foreignkey')
    op.drop_constraint('batch_column_fk_batch', 'batch_column', type_='foreignkey')

    # Rename all tables.
    op.rename_table('setting', 'settings')
    op.rename_table('change', 'changes')
    op.rename_table('phone', 'phone_numbers')
    op.rename_table('email', 'email_addresses')
    op.rename_table('person', 'people')
    op.rename_table('role', 'roles')
    op.rename_table('permission', 'permissions')
    op.rename_table('user', 'users')
    op.rename_table('user_x_role', 'users_roles')
    op.rename_table('store', 'stores')
    op.rename_table('employee', 'employees')
    op.rename_table('customer_group', 'customer_groups')
    op.rename_table('customer', 'customers')
    op.rename_table('customer_x_group', 'customers_groups')
    op.rename_table('customer_x_person', 'customers_people')
    op.rename_table('vendor', 'vendors')
    op.rename_table('vendor_contact', 'vendor_contacts')
    op.rename_table('department', 'departments')
    op.rename_table('subdepartment', 'subdepartments')
    op.rename_table('category', 'categories')
    op.rename_table('family', 'families')
    op.rename_table('report_code', 'report_codes')
    op.rename_table('brand', 'brands')
    op.rename_table('product', 'products')
    op.rename_table('product_code', 'product_codes')
    op.rename_table('product_cost', 'product_costs')
    op.rename_table('product_price', 'product_prices')
    op.rename_table('batch', 'batches')
    op.rename_table('batch_column', 'batch_columns')
    op.rename_table('label_profile', 'label_profiles')

    # Recreate all constraints.
    op.create_unique_constraint('roles_uq_name', 'roles', ['name'])
    op.create_foreign_key('permissions_fk_role', 'permissions', 'roles', ['role_uuid'], ['uuid'])
    op.create_foreign_key('users_fk_person', 'users', 'people', ['person_uuid'], ['uuid'])
    op.create_unique_constraint('users_uq_username', 'users', ['username'])
    op.create_foreign_key('users_roles_fk_role', 'users_roles', 'roles', ['role_uuid'], ['uuid'])
    op.create_foreign_key('users_roles_fk_user', 'users_roles', 'users', ['user_uuid'], ['uuid'])
    op.create_foreign_key('employees_fk_person', 'employees', 'people', ['person_uuid'], ['uuid'])
    op.create_foreign_key('customers_groups_fk_group', 'customers_groups', 'customer_groups', ['group_uuid'], ['uuid'])
    op.create_foreign_key('customers_groups_fk_customer', 'customers_groups', 'customers', ['customer_uuid'], ['uuid'])
    op.create_foreign_key('customers_people_fk_customer', 'customers_people', 'customers', ['customer_uuid'], ['uuid'])
    op.create_foreign_key('customers_people_fk_person', 'customers_people', 'people', ['person_uuid'], ['uuid'])
    op.create_unique_constraint('vendor_uq_id', 'vendors', ['id'])
    op.create_foreign_key('vendor_contacts_fk_vendor', 'vendor_contacts', 'vendors', ['vendor_uuid'], ['uuid'])
    op.create_foreign_key('vendor_contacts_fk_person', 'vendor_contacts', 'people', ['person_uuid'], ['uuid'])
    op.create_foreign_key('categories_fk_department', 'categories', 'departments', ['department_uuid'], ['uuid'])
    op.create_unique_constraint('report_codes_uq_code', 'report_codes', ['code'])
    op.create_foreign_key('products_fk_department', 'products', 'departments', ['department_uuid'], ['uuid'])
    op.create_foreign_key('products_fk_subdepartment', 'products', 'subdepartments', ['subdepartment_uuid'], ['uuid'])
    op.create_foreign_key('products_fk_category', 'products', 'categories', ['category_uuid'], ['uuid'])
    op.create_foreign_key('products_fk_family', 'products', 'families', ['family_uuid'], ['uuid'])
    op.create_foreign_key('products_fk_report_code', 'products', 'report_codes', ['report_code_uuid'], ['uuid'])
    op.create_foreign_key('products_fk_brand', 'products', 'brands', ['brand_uuid'], ['uuid'])
    op.create_foreign_key('products_regular_price_uuid_fkey', 'products', 'product_prices', ['regular_price_uuid'], ['uuid'])
    op.create_foreign_key('products_current_price_uuid_fkey', 'products', 'product_prices', ['current_price_uuid'], ['uuid'])
    op.create_index('products_ix_upc', 'products', ['upc'])
    op.create_foreign_key('product_codes_fk_product', 'product_codes', 'products', ['product_uuid'], ['uuid'])
    op.create_foreign_key('product_costs_fk_vendor', 'product_costs', 'vendors', ['vendor_uuid'], ['uuid'])
    op.create_foreign_key('product_costs_fk_product', 'product_costs', 'products', ['product_uuid'], ['uuid'])
    op.create_foreign_key('product_prices_fk_product', 'product_prices', 'products', ['product_uuid'], ['uuid'])
    op.create_foreign_key('batch_columns_fk_batch', 'batch_columns', 'batches', ['batch_uuid'], ['uuid'])



# -*- coding: utf-8 -*-
"""add product_image

Revision ID: 234c79420312
Revises: 7ba105af22a7
Create Date: 2017-02-22 22:43:58.290180

"""

from __future__ import unicode_literals

# revision identifiers, used by Alembic.
revision = '234c79420312'
down_revision = u'7ba105af22a7'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types
from sqlalchemy.dialects import postgresql


def upgrade():

    # product_image
    op.create_table('product_image',
                    sa.Column('uuid', sa.String(length=32), nullable=False),
                    sa.Column('product_uuid', sa.String(length=32), nullable=False),
                    sa.Column('bytes', sa.LargeBinary(), nullable=False),
                    sa.ForeignKeyConstraint(['product_uuid'], [u'product.uuid'], name=u'product_image_fk_product'),
                    sa.PrimaryKeyConstraint('uuid')
    )


def downgrade():

    # product_image
    op.drop_table('product_image')



# -*- coding: utf-8; -*-
"""misc. more things

Revision ID: 08dcf4eebe66
Revises: 4638ceca90d6
Create Date: 2021-03-09 18:53:22.103819

"""

from __future__ import unicode_literals, absolute_import

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '08dcf4eebe66'
down_revision = '4638ceca90d6'
branch_labels = None
depends_on = None


def upgrade():

    # transaction
    op.add_column('transaction', sa.Column('effective_date', sa.Date(), nullable=True))
    op.add_column('transaction', sa.Column('shopper_level_number', sa.Integer(), nullable=True))
    op.add_column('transaction', sa.Column('store_id', sa.String(length=10), nullable=True))

    # transaction_item
    op.add_column('transaction_item', sa.Column('subdepartment_name', sa.String(length=30), nullable=True))
    op.add_column('transaction_item', sa.Column('subdepartment_number', sa.Integer(), nullable=True))


def downgrade():

    # transaction_item
    op.drop_column('transaction_item', 'subdepartment_number')
    op.drop_column('transaction_item', 'subdepartment_name')

    # transaction
    op.drop_column('transaction', 'store_id')
    op.drop_column('transaction', 'shopper_level_number')
    op.drop_column('transaction', 'effective_date')



# -*- coding: utf-8 -*-

from __future__ import unicode_literals, absolute_import

import copy
from contextlib import contextmanager

from mock import patch

from rattail.tests import NullProgress


class ImporterTester(object):
    """
    Mixin for importer test suites.
    """
    handler_class = None
    importer_class = None
    sample_data = {}

    def make_handler(self, **kwargs):
        if 'config' not in kwargs and hasattr(self, 'config'):
            kwargs['config'] = self.config
        return self.handler_class(**kwargs)

    def make_importer(self, **kwargs):
        if 'config' not in kwargs and hasattr(self, 'config'):
            kwargs['config'] = self.config
        kwargs.setdefault('progress', NullProgress)
        return self.importer_class(**kwargs)

    def copy_data(self):
        return copy.deepcopy(self.sample_data)

    @contextmanager
    def host_data(self, data):
        self._host_data = data
        host_data = [self.importer.normalize_host_object(obj) for obj in data.values()]
        with patch.object(self.importer, 'normalize_host_data') as normalize:
            normalize.return_value = host_data
            yield

    @contextmanager
    def local_data(self, data):
        self._local_data = data
        local_data = {}
        for key, obj in data.items():
            normal = self.importer.normalize_local_object(obj)
            local_data[self.importer.get_key(normal)] = {'object': obj, 'data': normal}
        with patch.object(self.importer, 'cache_local_data') as cache:
            cache.return_value = local_data
            yield

    def import_data(self, host_data=None, local_data=None, **kwargs):
        if host_data is None:
            host_data = self.sample_data
        if local_data is None:
            local_data = self.sample_data
        with self.host_data(host_data):
            with self.local_data(local_data):
                self.result = self.importer.import_data(**kwargs)

    def assert_import_created(self, *keys):
        created, updated, deleted = self.result
        self.assertEqual(len(created), len(keys))
        for key in keys:
            key = self.importer.get_key(self._host_data[key])
            found = False
            for local_object, host_data in created:
                if self.importer.get_key(host_data) == key:
                    found = True
                    break
            if not found:
                raise self.failureException("Key {} not created when importing with {}".format(key, self.importer))

    def assert_import_updated(self, *keys):
        created, updated, deleted = self.result
        self.assertEqual(len(updated), len(keys))
        for key in keys:
            key = self.importer.get_key(self._host_data[key])
            found = False
            for local_object, local_data, host_data in updated:
                if self.importer.get_key(local_data) == key:
                    found = True
                    break
            if not found:
                raise self.failureException("Key {} not updated when importing with {}".format(key, self.importer))

    def assert_import_deleted(self, *keys):
        created, updated, deleted = self.result
        self.assertEqual(len(deleted), len(keys))
        for key in keys:
            key = self.importer.get_key(self._local_data[key])
            found = False
            for local_object, local_data in deleted:
                if self.importer.get_key(local_data) == key:
                    found = True
                    break
            if not found:
                raise self.failureException("Key {} not deleted when importing with {}".format(key, self.importer))
